@@ -1,17 +1,25 @@
+from __future__ import annotations
+from typing import List, TYPE_CHECKING
 import os.path, shutil
-import pathlib, stat, time, subprocess, concurrent.futures
-from mama.system import System, console, execute, execute_echo
-from mama.util import save_file_if_contents_changed, glob_with_name_match, \
-                      normalized_path, write_text_to, copy_if_needed
-from mama.build_dependency import BuildDependency
-from mama.build_config import BuildConfig
-from mama.papa_deploy import papa_deploy_to, Asset
+
+from .types.git import Git
+from .types.local_source import LocalSource
+from .types.asset import Asset
+from .types.artifactory_pkg import ArtifactoryPkg
+
+from .artifactory import artifactory_fetch_and_reconfigure
+from .utils.system import System, console
+from .utils.sub_process import execute, execute_echo
+from .papa_deploy import papa_deploy_to, papa_upload_to
+import mama.msbuild as msbuild
 import mama.util as util
-from typing import List
-from mama.msbuild import msbuild_build
 import mama.cmake_configure as cmake
 import mama.package as package
-import mama.add_dependencies as add_dependencies
+
+if TYPE_CHECKING:
+    from .build_config import BuildConfig
+    from .build_dependency import BuildDependency
+
 
 ######################################################################################
 
@@ -43,16 +51,17 @@ class BuildTarget:
             self.papa_deploy('deploy/MyProject')
     ```
     """
-    def __init__(self, name, config:BuildConfig, dep:BuildDependency, args:list):
+    def __init__(self, name, config:BuildConfig, dep:BuildDependency, args:List[str]):
         if config is None: raise RuntimeError(f'BuildTarget {name} config argument must be set')
         if dep is None:    raise RuntimeError(f'BuildTarget {name} dep argument must be set')
         self.config = config
         self.name = name
         self.dep  = dep
         self.args = [] # user defined args for this target (must be a list)
-        self.install_target   = 'install'
+        self.install_target = 'install'
         self.cmake_ndk_toolchain   = '' # Custom Android toolchain file for this target only
         self.cmake_raspi_toolchain = '' # Custom Raspberry toolchain file for this target only
+        self.cmake_oclea_toolchain = '' # Custom Oclea toolchain file for this target only
         self.cmake_ios_toolchain   = '' # Custom iOS toolchain file for this target only
         self.cmake_opts       = []
         self.cmake_cxxflags   = dict()
@@ -65,8 +74,9 @@ class BuildTarget:
         self.enable_fortran_build = False
         self.enable_cxx_build = True
         self.enable_multiprocess_build = True
-        self.gcc_clang_visibility_hidden = True # -fvisibility=hidden 
-        self.build_dependencies = [] # dependency files
+        self.clean_intermediate_files = True # delete .o and .obj files after build success if not root or always_build
+        self.gcc_clang_visibility_hidden = True # -fvisibility=hidden
+        self.build_products = [] # executables/libs products from last build
         self.no_includes = False # no includes to export
         self.no_libs = False # no libs to export
         self.exported_includes = [] # include folders to export from this target
@@ -79,19 +89,25 @@ class BuildTarget:
         self.ios     = self.config.ios
         self.android = self.config.android
         self.raspi   = self.config.raspi
+        self.oclea   = self.config.oclea
         self.os_windows = System.windows
         self.os_linux   = System.linux
         self.os_macos   = System.macos
         self._set_args(args)
 
 
-    def _set_args(self, args:list):
+    def _set_args(self, args:List[str]):
         if not isinstance(args, list):
             raise RuntimeError(f'BuildTarget {self.name} target args must be a list')
         self.args += args
         #console(f'Added args to {self.name}: {self.args}')
 
 
+    def children(self) -> List[BuildDependency]:
+        """ Get resolved child dependencies """
+        return self.dep.get_children()
+
+
     def source_dir(self, subpath=''):
         """
         Returns the current source directory.
@@ -102,6 +118,7 @@ class BuildTarget:
             # --> C:/Projects/ReCpp/lib/ReCpp.lib
         ```
         """
+        if not subpath: return self.dep.src_dir
         return util.path_join(self.dep.src_dir, subpath)
 
 
@@ -115,9 +132,34 @@ class BuildTarget:
             # --> C:/Projects/ReCpp/build/windows/lib/ReCpp.lib
         ```
         """
+        if not subpath: return self.dep.build_dir
         return util.path_join(self.dep.build_dir, subpath)
 
 
+    def set_artifactory_ftp(self, ftp_url, auth='store'):
+        """
+        Configures the remote Artifactory FTP URL where packages
+        will be checked for download. If a package with correct commit hash
+        exists, it will be used instead of building locally.
+
+        If auth='store' then system's secure keyring is used to store
+        the credentials. If authentication fails, then credentials are cleared.
+
+        The username and password can be overriden by ENV variables
+        `MAMA_ARTIFACTORY_USER` and `MAMA_ARTIFACTORY_PASS` for use in build systems
+        ```
+            def dependencies(self):
+                self.config.set_artifactory_url('myserver.com', auth='store')
+                self.config.set_artifactory_url('myserver.com', auth='prompt')
+        ```
+        NOTE: Currently only FTP is supported
+        """
+        if not self.dep.is_root:
+            console(f'WARNING: {self.name} set_artifactory_ftp ignored for non-root targets')
+            return
+        self.config.set_artifactory_ftp(ftp_url=ftp_url, auth=auth)
+
+
     def add_local(self, name, source_dir, mamafile=None, always_build=False, args=[]):
         """
         Add a local dependency. This can be a git submodule or just some local folder.
@@ -138,7 +180,7 @@ class BuildTarget:
         self.add_local('avdecoder', 'lib/avdecoder', always_build=True)
         ```
         """
-        add_dependencies.add_local(self, name, source_dir, mamafile, always_build, args)
+        self.dep.add_child(LocalSource(name, source_dir, mamafile, always_build, args))
 
 
     def add_git(self, name, git_url, git_branch='', git_tag='', mamafile=None, args=[]):
@@ -149,17 +191,42 @@ class BuildTarget:
     
         If the remote GIT repository does not contain a `mamafile.py`, you will have to
         provide your own relative or absolute mamafile path.
-    
+
+        For PUBLIC repositories, only use `https://` to prevent clone failures!!!
+
         Any arguments are passed onto child targets as `self.args`.
         ```
-        self.add_git('ReCpp', 'https://github.com/RedFox20/ReCpp.git')
-        self.add_git('ReCpp', 'https://github.com/RedFox20/ReCpp.git', 
-                     git_branch='master')
+        self.add_git('ReCpp', 'git@github.com:RedFox20/ReCpp.git')
+        self.add_git('ReCpp', 'git@github.com:RedFox20/ReCpp.git', git_branch='master')
         self.add_git('opencv', 'https://github.com/opencv/opencv.git', 
                      git_branch='3.4', mamafile='mama/opencv_cfg.py')
         ```
         """
-        add_dependencies.add_git(self, name, git_url, git_branch, git_tag, mamafile, args)
+        self.dep.add_child(Git(name, git_url, git_branch, git_tag, mamafile, args))
+
+
+    def add_artifactory_pkg(self, name, version='latest', fullname=None):
+        """
+        Adds an Artifactory only dependency.
+        The dependency will be downloaded from the artifactory url.
+
+        If the remote artifactory does not contain this package,
+        an error is thrown during build.
+
+        If a version value is given, mamabuild will try to automatically
+        figure out the appropriate remote package.
+
+        If a fullname value is given, only the specific artifactory package
+        will be used as an override. This is mostly useful for source-only packages
+        and for platform-specific configuration.
+
+        ```
+        self.add_artifactory_pkg('mylib', version='latest')
+        self.add_artifactory_pkg('mylib', version='df76b66')
+        self.add_artifactory_pkg('mylib', fullname='mylib-linux-x64-release-df76b66')
+        ```
+        """
+        self.dep.add_child(ArtifactoryPkg(name, version=version, fullname=fullname))
 
 
     def get_dependency(self, name):
@@ -171,7 +238,7 @@ class BuildTarget:
         """
         if self.dep.name == name:
             return self.dep
-        for dep in self.dep.children:
+        for dep in self.children():
             if dep.name == name:
                 return dep
         raise KeyError(f"BuildTarget {self.name} has no child dependency named '{name}'")
@@ -186,12 +253,13 @@ class BuildTarget:
         """
         if self.name == name:
             return self
-        for dep in self.dep.children:
+        for dep in self.children():
             if dep.name == name:
                 return dep.target
         raise KeyError(f"BuildTarget {self.name} has no child target named '{name}'")
 
 
+    ## TODO: Move this into `package.py`
     def inject_products(self, dst_dep, src_dep, include_path, libs, libfilters=None):
         """
         Injects products from `src_dep` into `dst_dep` as CMake defines.
@@ -213,10 +281,10 @@ class BuildTarget:
         """
         dst_dep = self.get_dependency(dst_dep)
         src_dep = self.get_dependency(src_dep)
-        dst_dep.depends_on.append(src_dep)
         dst_dep.product_sources.append( (src_dep, include_path, libs, libfilters) )
 
 
+    ## TODO: Move this into `package.py`
     def get_product_defines(self):
         """
         Collects all results injected by `inject_products()`.
@@ -287,9 +355,9 @@ class BuildTarget:
         """
         dependency = all if all else self.select(windows, linux, macos, ios, android)
         if dependency:
-            dependency = normalized_path(os.path.join(self.dep.build_dir, dependency))
-            self.build_dependencies.append(dependency)
-            #console(f'    {self.name}.build_dependencies += {dependency}')
+            dependency = util.normalized_join(self.build_dir(), dependency)
+            self.build_products.append(dependency)
+            #console(f'    {self.name}.build_products += {dependency}')
 
 
     def no_export_includes(self):
@@ -487,7 +555,7 @@ class BuildTarget:
         for flag in flags:
             if isinstance(flag, list): self.add_cxx_flags(*flag)
             else: self._add_dict_flag(self.cmake_cxxflags, flag)
-    
+
 
     def add_c_flags(self, *flags):
         """
@@ -630,8 +698,8 @@ class BuildTarget:
                       self.source_dir('deploy/Awesome.aar/jni/armeabi-v7a'))
         ```
         """
-        console(f'copy {src} --> {dst}')
-        copy_if_needed(src, dst)
+        if self.config.verbose: console(f'copy {src} --> {dst}')
+        util.copy_if_needed(src, dst)
 
 
     def copy_built_file(self, builtFile, copyToFolder):
@@ -641,8 +709,10 @@ class BuildTarget:
             self.copy_built_file('RelWithDebInfo/libawesome.a', 'lib')
         ```
         """
-        src = f'{self.dep.build_dir}/{builtFile}'
-        dst = f'{self.dep.build_dir}/{copyToFolder}'
+        src = f'{self.build_dir()}/{builtFile}'
+        dst = f'{self.build_dir()}/{copyToFolder}/{os.path.basename(builtFile)}'
+        if not os.path.exists(src) and os.path.exists(dst):
+            return # src is missing, but dst exists, ignore error
         shutil.copy(src, dst)
 
 
@@ -654,7 +724,7 @@ class BuildTarget:
             # --> 'C:/Projects/Game/Plugins/NanoMesh
         ```
         """
-        copy_if_needed(self.source_dir(src_dir), dst_dir)
+        util.copy_if_needed(self.source_dir(src_dir), dst_dir)
 
 
     def download_file(self, remote_url, local_dir, force=False):
@@ -665,7 +735,7 @@ class BuildTarget:
             # --> 'bin/file1'
         ```
         """
-        util.download_file(remote_url, local_dir, force)
+        return util.download_file(remote_url, local_dir, force)
 
 
     def download_and_unzip(self, remote_zip, extract_dir, unless_file_exists=None):
@@ -676,9 +746,11 @@ class BuildTarget:
         ```
             self.download_and_unzip('http://example.com/archive.zip', 
                                     'bin', 'bin/unzipped_file.txt')
+            # --> 'bin/'  on success
+            # --> None    on failure
         ```
         """
-        util.download_and_unzip(remote_zip, extract_dir, unless_file_exists)
+        return util.download_and_unzip(remote_zip, extract_dir, unless_file_exists)
 
 
     def visibility_hidden(self, hidden=True):
@@ -750,7 +822,7 @@ class BuildTarget:
             self.run('make release -j7')
         ```
         """
-        dir = self.dep.src_dir if src_dir else self.dep.build_dir
+        dir = self.source_dir() if src_dir else self.build_dir()
         execute(f'cd {dir} && {command}', echo=True)
 
 
@@ -764,7 +836,8 @@ class BuildTarget:
         """
         execute_echo(working_dir, command)
 
-    
+
+    ## TODO: Move this into a new utility
     def gdb(self, command, src_dir=True):
         """
         Run a command with gdb in the build folder.
@@ -772,14 +845,14 @@ class BuildTarget:
             self.gdb('bin/NanoMeshTests')
         ```
         """
-        if self.android or self.ios or self.raspi:
-            console('Cannot run tests for Android, iOS, Raspi builds.')
+        if self.android or self.ios or self.raspi or self.oclea:
+            console('Cannot run tests for Android, iOS, Raspi, Oclea builds.')
             return # nothing to run
-        
+
         split = command.split(' ', 1)
         cmd = split[0].lstrip('.')
         args = split[1] if len(split) >= 2 else ''
-        path = self.dep.src_dir if src_dir else self.dep.build_dir
+        path = self.source_dir() if src_dir else self.build_dir()
         path = f"{path}/{os.path.dirname(cmd).lstrip('/')}"
         exe = os.path.basename(cmd)
 
@@ -805,14 +878,19 @@ class BuildTarget:
 
     def dependencies(self):
         """
-        Add any dependencies in this step.
+        Add any additional dependencies in this step,
+        or setup project configuration for root targets.
+
+        If this target is fetched as a package from the artifactory,
+        then this step is skipped!
         ```
-        class MyProject(mama.BuildTarget):
+        class MyRootProject(mama.BuildTarget):
             def dependencies(self):
-                self.add_git('ReCpp',
-                             'http://github.com/RedFox20/ReCpp.git')
-                self.add_local('fbxsdk',
-                               'third_party/FBX')
+                # only valid for root targets
+                self.set_artifactory_ftp('artifacts.myftp.com', auth='store')
+
+                self.add_git('ReCpp', 'http://github.com/RedFox20/ReCpp.git')
+                self.add_local('fbxsdk', 'third_party/FBX')
         ```
         """
         pass
@@ -895,8 +973,9 @@ class BuildTarget:
         """
         if self.no_includes: self.default_package_includes()
         if self.no_libs: self.default_package_libs()
-    
 
+
+    ## TODO: move this into `package.py`
     def default_package_includes(self):
         """
         Performs default INCLUDE packaging steps.
@@ -913,6 +992,7 @@ class BuildTarget:
         elif self.export_include('',        build_dir=False): pass
 
 
+    ## TODO: move this into `package.py`
     def default_package_libs(self):
         """
         Performs default LIB packaging steps.
@@ -942,15 +1022,15 @@ class BuildTarget:
         ```
         """
         self.default_deploy()
-    
+
 
     def default_deploy(self):
-        self.papa_deploy(f'deploy/{self.name}')
+        self.papa_deploy(f'deploy/{self.name}', src_dir=False)
 
-    
-    def papa_deploy(self, package_path, src_dir=True,
-                    r_includes=True, r_dylibs=True,
-                    r_syslibs=False, r_assets=True):
+
+    def papa_deploy(self, package_path, src_dir=False,
+                    r_includes=False, r_dylibs=False,
+                    r_syslibs=False, r_assets=False):
         """
         This will create a PAPA package, which includes
             package_path/papa.txt
@@ -971,7 +1051,7 @@ class BuildTarget:
             MyPackageName/libawesome.so
             MyPackageName/include/...
             MyPackageName/someassets/extra.txt
-            
+        
         PAPA descriptor `papa.txt` format:
             P MyPackageName
             I include
@@ -979,8 +1059,12 @@ class BuildTarget:
             S libGL.a
             A someassets/extra.txt
         """
-        path = package.target_root_path(self, package_path, src_dir)
-        papa_deploy_to(self, path, r_includes, r_dylibs, r_syslibs, r_assets)
+        if self.config.list:
+            return # don't deploy during listing
+        self.papa_path = package.target_root_path(self, package_path, src_dir)
+        papa_deploy_to(self, self.papa_path, \
+            r_includes=r_includes, r_dylibs=r_dylibs, \
+            r_syslibs=r_syslibs, r_assets=r_assets)
 
 
     def test(self, args):
@@ -1029,18 +1113,35 @@ class BuildTarget:
             console(f"CMakeBuild {self.name}  ({self.cmake_build_type})")
         self.dep.ensure_cmakelists_exists()
         cmake.inject_env(self)
-        cmake.run_config(self)
-        cmake.run_build(self, install=True)
+        cmake.run_config(self) # THROWS on CMAKE failure
+        cmake.run_build(self, install=True) # THROWS on CMAKE failure
 
 
     def is_test_target(self):
         """
-        TRUE if this build target was specified along with test command:
-        `mama test target=this_name`
+        TRUE if this build target was specified along with `test` command.
+        This matches `all`, specific cmdline targets, and the `root` target.
+        ```
+        mama test              # the root target
+        mama test this_target  # specific target
+        mama test all          # all targets
+        ```
         """
         return self.config.test and self.dep.is_root_or_config_target()
 
 
+    def is_current_target(self):
+        """
+        TRUE if this BuildTarget is a configuration target for
+        build/test/etc. This matches 'all' or specific cmdline targets:
+        ```
+            mama build
+            mama build this_target
+        ```
+        """
+        return self.config.target_matches(self.name)
+
+
     ## Build only this target
     def _execute_tasks(self):
         if self.dep.already_executed:
@@ -1052,44 +1153,74 @@ class BuildTarget:
             self._execute_run_tasks()
         except Exception as err:
             import traceback
-            tb = traceback.format_exc().splitlines(True)
-            tb = tb[-3]+tb[-2] if len(tb) >= 3 else ''.join(tb)
-            console(f'  [BUILD FAILED]  {self.dep.name}  \n{err}\n\nError Source:\n{tb}')
-            exit(-1) # exit without stack trace
+            console(f'  [BUILD FAILED]  {self.dep.name}  \n{err}\n\n')
+            traceback.print_exc()
+            exit(-1) # exit without further stack trace
 
 
+    def try_automatic_artifactory_fetch(self):
+        is_deploy = self.config.deploy or self.config.upload
+        is_target = self.is_current_target()
+        # auto-fetch if it's not a deploy or if we're not the target
+        if not is_deploy or not is_target:
+            fetched, _ = artifactory_fetch_and_reconfigure(self) # this will reconfigure packaging
+            return fetched
+        return None
+
+
+    ## TODO: move all of this into a new utility
     def _execute_build_tasks(self):
-        if self.dep.should_rebuild and not self.dep.nothing_to_build:
+        can_build = not self.dep.nothing_to_build
+        if can_build and self.dep.should_rebuild:
             self.configure() # user customization
-            if not self.dep.nothing_to_build:
-                self.build() # user customization
-                self.dep.successful_build()
-        
-        self.package() # user customization
+            if can_build:
+                fetched = self.try_automatic_artifactory_fetch()
+                if not fetched:
+                    self.build() # user build customization
 
-        # no packaging provided by user; use default packaging instead
-        if not self.exported_includes and not self.no_includes:
-            self.default_package_includes()
-        if not (self.exported_libs or self.exported_syslibs) and not self.no_libs:
-            self.default_package_libs()
+                self.dep.successful_build()
+
+                # NOTE: clean_intermediate_files is a suggestion !
+                # for `always_build` and `root` we don't want to clean the files
+                if fetched or \
+                    (self.clean_intermediate_files and not (self.dep.always_build or self.dep.is_root)):
+                    package.clean_intermediate_files(self)
+
+        # skip package() if we already fetched it as a package from artifactory()
+        if not self.dep.from_artifactory:
+            self.package() # user customization
+
+            # no packaging provided by user; use default packaging instead
+            if not self.exported_includes and not self.no_includes:
+                self.default_package_includes()
+            if not (self.exported_libs or self.exported_syslibs) and not self.no_libs:
+                self.default_package_libs()
 
         # only save and print exports if we built anything
         if self.dep.build_dir_exists():
             self.dep.save_exports_as_dependencies(self.exported_libs)
             # print exports only if target match
-            if self.config.target_matches(self.name):
-                self._print_exports()
+            if self.is_current_target():
+                self.print_exports()
 
 
     def _execute_deploy_tasks(self):
-        # Only run deploy for either:
-        # -> Root target by default
-        # -> or Specific target
-        if self.config.deploy:
-            specific_target = not self.config.no_specific_target()
-            if (not specific_target and self.dep.is_root) or \
-                (specific_target and self.config.target_matches(self.name)):
-                self.deploy() # user customization
+        if not self.config.deploy and not self.config.upload:
+            return
+
+        no_targets = not self.config.target and self.dep.is_root # only root target
+        for_all = self.config.targets_all() # all targets
+        one_target = not for_all and self.is_current_target() # only one target
+        if not (for_all or no_targets or one_target):
+            return # not going to deploy
+
+        self.deploy() # user customization
+
+        if self.config.upload:
+            if not self.papa_path:
+                raise RuntimeError(f'BuildTarget {self.name} was not deployed! '\
+                                    'Add self.papa_deploy() to mamafile deploy()!')
+            papa_upload_to(self, self.papa_path)
 
 
     def _execute_run_tasks(self):
@@ -1104,32 +1235,32 @@ class BuildTarget:
             self.start(start_args)
 
 
-    def _print_ws_path(self, what, path, check_exists=True):
+    def _print_ws_path(self, what, path, abs_path, check_exists=True):
         def exists():
             return '' if os.path.exists(path) else '   !! (path does not exist) !!'
         if path.startswith('-framework'):
             console(f'    {what}  {path}')
-        elif path.startswith(self.config.workspaces_root):
+        elif not abs_path and path.startswith(self.config.workspaces_root):
             console(f'    {what}  {path[len(self.config.workspaces_root) + 1:]}{exists()}')
-        elif path.startswith(self.dep.src_dir):
-            console(f'    {what}  {path[len(self.dep.src_dir) + 1:]}{exists()}')
+        elif not abs_path and path.startswith(self.source_dir()):
+            console(f'    {what}  {path[len(self.source_dir()) + 1:]}{exists()}')
         else:
             ex = exists() if check_exists else ''
             console(f'    {what}  {path}{ex}')
 
 
-    def _print_exports(self):
+    def print_exports(self, abs_paths=False):
         if not self.config.print:
             return
-        if not (self.exported_includes and self.exported_libs and self.exported_syslibs and self.exported_assets):
+        if not (self.exported_includes or self.exported_libs or self.exported_syslibs or self.exported_assets):
             return
 
         console(f'  - Package {self.name}')
-        for include in self.exported_includes: self._print_ws_path('<I>', include)
-        for library in self.exported_libs:     self._print_ws_path('[L]', library)
-        for library in self.exported_syslibs:  self._print_ws_path('[S]', library, check_exists=False)
-        if self.config.deploy:
-            for asset in self.exported_assets: self._print_ws_path('[A]', str(asset), check_exists=False)
+        for include in self.exported_includes: self._print_ws_path('<I>', include, abs_paths)
+        for library in self.exported_libs:     self._print_ws_path('[L]', library, abs_paths)
+        for library in self.exported_syslibs:  self._print_ws_path('[S]', library, abs_paths, check_exists=False)
+        if self.config.deploy or self.config.upload:
+            for asset in self.exported_assets: self._print_ws_path('[A]', str(asset), abs_paths, check_exists=False)
         elif self.exported_assets:
             assets = 'assets' if len(self.exported_assets) > 1 else 'asset'
             console(f'    [A]  ({len(self.exported_assets)} {assets})')
@@ -1158,7 +1289,7 @@ class BuildTarget:
         if self.config.print:
             console('\n#########################################')
             console(f'MSBuild {self.name} {projectfile}')
-        msbuild_build(self.config, self.source_dir(projectfile), properties)
+        msbuild.msbuild_build(self.config, self.source_dir(projectfile), properties)
 
 
 ######################################################################################
