@@ -1,56 +1,41 @@
-import os, subprocess, shlex, time
-from mama.system import System, console, execute
-from mama.build_config import BuildConfig
-from mama.async_file_reader import AsyncFileReader, AsyncConsoleReader
+import os
+from .utils.system import System, console
+from .build_config import BuildConfig
+from .utils.sub_process import SubProcess, execute
 
 
-def _rerunnable_cmake_conf(cwd, args, allow_rerun, target):
+def _rerunnable_cmake_conf(cmd, cwd, allow_rerun, target):
     rerun = False
     error = ''
     delete_cmakecache = False
     print_enabled = target.config.print
     verbose = target.config.verbose
-    if verbose: console(args)
+    if verbose: console(cmd)
     #xcode_filter = (target.ios or target.macos) and not target.enable_ninja_build 
-    proc = subprocess.Popen(args, shell=True, universal_newlines=True, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-    reader = AsyncConsoleReader(proc.stdout, proc.stderr)
 
-    def handle_error(err):
+    def handle_output(line):
         nonlocal rerun, delete_cmakecache
-        print(err, flush=True, end='')
-        if err.startswith('CMake Error: The source'):
+        print(line) # newline is not included
+        if line.startswith('CMake Error: The source'):
             rerun = True
             delete_cmakecache = True
         elif System.windows:
             # this happens every time MSVC compiler is updated. simple fix is to rerun cmake
-            rerun |= err.startswith('  is not a full path to an existing compiler tool.')
-        elif err.startswith('CMake Error: Error: generator :') or \
-             err.startswith('CMake Error: The source'):
+            rerun |= line.startswith('  is not a full path to an existing compiler tool.')
+        elif line.startswith('CMake Error: Error: generator :') or \
+             line.startswith('CMake Error: The source'):
             rerun = True
             delete_cmakecache = True
 
-    def read_out_err():
-        while reader.available():
-            out, err = reader.read()
-            if out:
-                if print_enabled:
-                    print(out, flush=True, end='')
-            elif err:
-                handle_error(err)
+    exit_status = SubProcess.run(cmd, cwd, io_func=handle_output)
 
-    while proc.poll() is None:
-        read_out_err()
-        time.sleep(0.015)
-    reader.stop()
-    read_out_err()
-    
     if delete_cmakecache:
         if print_enabled: console('Deleting CMakeCache.txt')
         os.remove(target.build_dir('CMakeCache.txt'))
     if rerun and allow_rerun:
         if print_enabled: console('Rerunning CMake configure')
-        return _rerunnable_cmake_conf(cwd, args, False, target)
-    if proc.returncode != 0:
+        return _rerunnable_cmake_conf(cmd, cwd, False, target)
+    if exit_status != 0:
         raise Exception(f'CMake configure error: {error}')
 
 
@@ -63,13 +48,13 @@ def run_config(target):
     
     cmake_flags = get_flags()
     generator = _generator(target)
-    src_dir = target.dep.src_dir
+    src_dir = target.source_dir()
     cmd = f'cmake {generator} {cmake_flags} -DCMAKE_INSTALL_PREFIX="." "{src_dir}"'
-    _rerunnable_cmake_conf(target.dep.build_dir, cmd, True, target)
+    _rerunnable_cmake_conf(cmd, target.build_dir(), True, target)
 
 
 def run_build(target, install, extraflags=''):
-    build_dir = target.dep.build_dir
+    build_dir = target.build_dir()
     flags = _build_config(target, install)
     extraflags = _buildsys_flags(target)
     execute(f'cmake --build {build_dir} {flags} {extraflags}', echo=target.config.verbose)
@@ -83,6 +68,7 @@ def _generator(target):
     if config.android:            return '-G "Unix Makefiles"'
     if config.linux:              return '-G "Unix Makefiles"'
     if config.raspi:              return '-G "Unix Makefiles"'
+    if config.oclea:              return '-G "Unix Makefiles"'
     if config.ios:                return '-G "Xcode"'
     if config.macos:              return '-G "Xcode"'
     else:                         return ''
@@ -103,7 +89,7 @@ def _make_program(target):
 
 def _custom_compilers(target):
     config:BuildConfig = target.config
-    cc, cxx = config.get_preferred_compiler_paths(target.enable_cxx_build)
+    cc, cxx, ver = config.get_preferred_compiler_paths()
     compilers = [f'CMAKE_C_COMPILER={cc}']
     if target.enable_cxx_build:
         compilers.append(f'CMAKE_CXX_COMPILER={cxx}')
@@ -167,12 +153,16 @@ def _default_options(target):
         add_flag('--sysroot', config.raspi_sysroot())
         for path in config.raspi_includes():
             add_flag(f'-I {path}')
+    elif config.oclea:
+        add_flag('--sysroot', config.oclea_sysroot())
+        for path in config.oclea_includes():
+            add_flag(f'-I {path}')
 
     if config.flags:
         add_flag(config.flags)
 
     opt = ["CMAKE_POSITION_INDEPENDENT_CODE=ON"]
-    if config.linux or config.raspi:
+    if config.linux or config.raspi or config.oclea:
         opt += _custom_compilers(target)
     
     if target.enable_fortran_build and config.fortran:
@@ -236,6 +226,20 @@ def _default_options(target):
             toolchain = target.source_dir(target.cmake_raspi_toolchain)
             if config.print: console(f'Toolchain: {toolchain}')
             opt += [f'CMAKE_TOOLCHAIN_FILE="{toolchain}"']
+    elif config.oclea:
+        opt += [
+            'OCLEA=TRUE',
+            'CMAKE_SYSTEM_NAME=Linux',
+            'CMAKE_SYSTEM_VERSION=1',
+            'CMAKE_SYSTEM_PROCESSOR=arm64',
+            'CMAKE_FIND_ROOT_PATH_MODE_PROGRAM=NEVER', # Use our definitions for compiler tools
+            'CMAKE_FIND_ROOT_PATH_MODE_LIBRARY=ONLY', # Search for libraries and headers in the target directories only
+            'CMAKE_FIND_ROOT_PATH_MODE_INCLUDE=ONLY',
+        ]
+        if target.cmake_oclea_toolchain:
+            toolchain = target.source_dir(target.cmake_oclea_toolchain)
+            if config.print: console(f'Toolchain: {toolchain}')
+            opt += [f'CMAKE_TOOLCHAIN_FILE="{toolchain}"']
     elif config.macos:
         pass
     elif config.ios:
