@@ -1,11 +1,14 @@
 #!/usr/bin/python3.6
-import sys, os, importlib
-from .system import console, execute
+import sys, os
+
+from .types.local_source import LocalSource
+from .utils.system import console
+from .utils.sub_process import execute
 from .util import glob_with_extensions, glob_folders_with_name_match
 from .build_config import BuildConfig
 from .build_target import BuildTarget
 from .build_dependency import BuildDependency
-from .dependency_chain import load_dependency_chain, execute_task_chain, find_dependency
+from .dependency_chain import load_dependency_chain, execute_task_chain, find_dependency, get_full_flattened_deps
 from .init_project import mama_init_project
 
 def print_title():
@@ -15,6 +18,7 @@ def print_usage():
     console('mama [actions...] [args...]')
     console('  actions:')
     console('    init       - create initial mamafile.py and CMakeLists.txt')
+    console('    list       - list all mama dependencies on this project')
     console('    build      - configure and build main project or specific target, this can clone, but does not pull')
     console('    update     - update and build target dependencies after calling git pull')
     console('    deploy     - runs PAPA deploy stage by gathering all libraries and assets')
@@ -23,6 +27,8 @@ def print_usage():
     console('    rebuild    - clean, update and build main project or specific target')
     console('    reclone    - wipe specific target dependency and clone it again')
     console('    wipe       - alias of reclone')
+    console('    upload     - uploads target package to artifactory server')
+    console('    if_needed  - only uploads if package does not exist on server')
     console('    test       - run tests for main project or specific target')
     console('    start=arg  - start a specific tool via mamafile.start(args)')
     console('    add        - add new dependency')
@@ -31,6 +37,7 @@ def print_usage():
     console('    help       - shows this help list')
     console('  install utils:')
     console('    install-clang6  - configures and installs clang6 for linux')
+    console('    install-clang11 - configures and installs clang11 for linux')
     console('    install-msbuild - configures and installs MSBuild for linux')
     console('  args:')
     console('    windows    - build for windows')
@@ -66,14 +73,17 @@ def print_usage():
     console('    mama update dep1               Update and build the specified target.')
     console('    mama serve android             Update, build and deploy for Android')
     console('    mama wipe dep1                 Wipe target dependency completely and clone again.')
+    console('    mama upload dep1               Deploys and uploads dependency to Artifactory server.')
     console('    mama test                      Run tests on main project.')
     console('    mama test=arg                  Run tests on main project with an argument.')
     console('    mama test="arg1 arg2"          Run tests on main project with multiple arguments.')
     console('    mama test dep1                 Run tests on target dependency project.')
     console('    mama start=dbtool              Call main project mamafile start() with args [`dbtool`].')
     console('  environment:')
-    console('    setenv("NINJA")                Path to NINJA build executable')
-    console('    setenv("ANDROID_HOME")         Path to Android SDK if auto-detect fails')
+    console('    setenv("NINJA")                  Path to NINJA build executable')
+    console('    setenv("ANDROID_HOME")           Path to Android SDK if auto-detect fails')
+    console('    setenv("MAMA_ARTIFACTORY_USER")  Username for Artifactory server')
+    console('    setenv("MAMA_ARTIFACTORY_PASS")  Password for Artifactory server')
 
 
 def open_project(config: BuildConfig, root_dependency: BuildDependency):
@@ -84,22 +94,28 @@ def open_project(config: BuildConfig, root_dependency: BuildDependency):
     
     if config.windows:
         solutions = glob_with_extensions(found.build_dir, ['.sln'])
-        if not solutions:
-            raise EnvironmentError('Could not find any Visual Studio solutions!')
-        execute(f'start {solutions[0]}', echo=True)
+        if solutions:
+            execute(f'start {solutions[0]}', echo=True)
+        else:
+            console('Could not find any Visual Studio solutions, using VSCode instead.')
+            execute(f'code {found.src_dir}', echo=True)
 
     elif config.macos or config.ios:
         projects = glob_folders_with_name_match(found.build_dir, ['.xcodeproj'])
-        if not projects:
-            raise EnvironmentError('Could not find any Xcode projects!')
-        execute(f'open {projects[0]}', echo=True)
+        if projects:
+            execute(f'open {projects[0]}', echo=True)
+        else:
+            console('Could not find any Xcode projects, using VSCode instead.')
+            execute(f'code {found.src_dir}', echo=True)
 
     elif config.linux:
-        raise EnvironmentError('Linux IDE selection not implemented. Try opening this folder with CLion.')
+        console(f'Using VSCode. You can also try opening this folder with CLion: {found.src_dir}')
+        execute(f'code {found.src_dir}', echo=True)
         #execute(f'xdg-open', echo=True)
 
     elif config.android:
-        raise EnvironmentError('Android IDE selection not implemented. Try opening this folder with Android Studio.')
+        console('Android IDE selection not implemented, using VSCode instead.')
+        execute(f'code {found.src_dir}', echo=True)
 
 def set_target_from_unused_args(config: BuildConfig):
     for arg in config.unused_args:
@@ -116,6 +132,7 @@ def check_config_target(config: BuildConfig, root: BuildDependency):
             console(f"ERROR: specified target='{config.target}' not found!")
             exit(-1)
 
+
 def main():
     if sys.version_info < (3, 6):
         console('FATAL ERROR: MamaBuild requires Python 3.6')
@@ -132,9 +149,15 @@ def main():
 
     source_dir = os.getcwd()
     name = os.path.basename(source_dir)
-    root = BuildDependency(name, config, BuildTarget, src=source_dir, is_root=True)
+    local_src = LocalSource(name, source_dir, mamafile=None, always_build=False, args=[])
+    workspace = None # figure out the workspace from the root mamafile.py
+    root = BuildDependency(None, config, workspace, local_src)
 
-    if config.mama_init:
+    if config.unused_args:
+        set_target_from_unused_args(config)
+
+    # root init
+    if config.mama_init and not config.target:
         mama_init_project(root)
         return
 
@@ -147,9 +170,6 @@ def main():
         console('FATAL ERROR: mamafile.py not found and CMakeLists.txt not found')
         exit(-1)
 
-    if config.unused_args:
-        set_target_from_unused_args(config)
-
     if config.update:
         if not config.target:
             config.target = 'all'
@@ -167,8 +187,31 @@ def main():
     load_dependency_chain(root)
     check_config_target(config, root)
 
+    # target init
+    if config.mama_init and config.target:
+        dep = find_dependency(root, config.target)
+        if not dep:
+            console(f'init command failed: target {config.target} not found')
+            exit(-1)
+        mama_init_project(dep)
+        return
+
+    if config.list:
+        console(f'    Dependency List: {get_full_flattened_deps(root)}')
+        if config.target:
+            dep = find_dependency(root, config.target)
+            if dep:
+                target:BuildTarget = dep.target
+                if dep.from_artifactory or target.try_automatic_artifactory_fetch():
+                    console(f'    Target {target.name} fetched from artifactory')
+                else:
+                    target.package()
+                target.print_exports(abs_paths=True)
+        return
+
     if config.android: config.init_ndk_path()
     if config.raspi:   config.init_raspi_path()
+    if config.oclea:   config.init_oclea_path()
 
     execute_task_chain(root)
 
