@@ -1,5 +1,6 @@
-import os, shutil, random, shlex, time, subprocess, pathlib, ssl, urllib, zipfile
-from mama.system import System, console, execute
+import os, shutil, random, pathlib, ssl, zipfile
+from .utils.system import System, console
+from .utils.sub_process import execute
 from urllib import request
 
 
@@ -8,6 +9,13 @@ def is_file_modified(src, dst):
            os.path.getsize(src) == os.path.getsize(dst)
 
 
+def find_executable_from_system(name):
+    if not name: return ''
+    output = shutil.which(name)
+    if not output: return ''
+    return output if os.path.isfile(output) else ''
+
+
 def copy_files(fromFolder, toFolder, fileNames):
     for file in fileNames:
         sourceFile = os.path.join(fromFolder, file)
@@ -42,26 +50,6 @@ def deploy_framework(framework, deployFolder):
     return False
 
 
-def run_with_timeout(executable, argstring, workingDir, timeoutSeconds=None):
-    args = [executable]
-    args += shlex.split(argstring)
-    start = time.time()
-    proc = subprocess.Popen(args, shell=True, cwd=workingDir)
-    try:
-        proc.wait(timeout=timeoutSeconds)
-        console(f'{executable} elapsed: {round(time.time()-start, 1)}s')
-    except subprocess.TimeoutExpired:
-        console('TIMEOUT, sending break signal')
-        if System.windows:
-            proc.send_signal(subprocess.signal.CTRL_C_EVENT)
-        else:
-            proc.send_signal(subprocess.signal.SIGINT)
-        raise
-    if proc.returncode == 0:
-        return
-    raise subprocess.CalledProcessError(proc.returncode, ' '.join(args))
-
-
 def has_contents_changed(filename, new_contents):
     if not os.path.exists(filename):
         return True
@@ -74,8 +62,8 @@ def save_file_if_contents_changed(filename, new_contents):
     pathlib.Path(filename).write_text(new_contents)
 
 
-# always join with forward slash /
 def path_join(first, second):
+    """ Always join with forward/ slashes """
     first  = first.rstrip('/\\')
     second = second.lstrip('/\\')
     if not first: return second
@@ -84,18 +72,26 @@ def path_join(first, second):
 
 
 def forward_slashes(pathstring):
+    """ Replace all back\\ slashes with forward/ slashes"""
     return pathstring.replace('\\', '/')
 
 
 def back_slashes(pathstring):
+    """ Replace all forward/ slashes with back\\ slashes"""
     return pathstring.replace('/', '\\')
 
 
 def normalized_path(pathstring):
+    """ Normalizes a path to ABSOLUTE path and all forward/ slashes """
     pathstring = os.path.abspath(pathstring)
     return pathstring.replace('\\', '/').rstrip()
 
 
+def normalized_join(path1, *pathsN):
+    """ Joins N paths and the calls normalized_path() """
+    return normalized_path(os.path.join(path1, *pathsN))
+
+
 def glob_with_extensions(rootdir, extensions):
     results = []
     for dirpath, _, dirfiles in os.walk(rootdir):
@@ -171,48 +167,67 @@ def read_lines_from(file):
         return f.readlines()
 
 
-def download_file(remote_url, local_dir, force=False):
+def get_file_size_str(size):
+    """
+    Returns file size as a human readable string, eg 96.5KB, or 100.1MB
+    """
+    if size < 128: return f'{size}B' # only show bytes for really small < 0.1 KB sizes
+    if size < (1024*1024): return f'{size/1024:.1f}KB'
+    if size < (1024*1024*1024): return f'{size/(1024*1024):.1f}MB'
+    return f'{size/(1024*1024):.2}GB'
+
+
+def download_file(remote_url, local_dir, force=False, message=None):
     local_file = os.path.join(local_dir, os.path.basename(remote_url))
     if not force and os.path.exists(local_file): # download file?
-        console(f"Using locally cached {local_file}")
+        console(f"    Using locally cached {local_file}")
         return local_file
     if not os.path.exists(local_dir):
         os.makedirs(local_dir, exist_ok=True)
+
     ctx = ssl.create_default_context()
     ctx.check_hostname = False
     ctx.verify_mode = ssl.CERT_NONE
+
     with request.urlopen(remote_url, context=ctx) as urlfile:
         with open(local_file, 'wb') as output:
-            total = int(urlfile.info()['Content-Length'].strip())
-            total_megas = int(total/(1024*1024))
-            prev_progress = -100
-            written = 0
+            size = int(urlfile.info()['Content-Length'].strip())
+            if not message: message = f'Downloading {remote_url}'
+            print(f'{message} {get_file_size_str(size)}')
+            print(f'    |{" ":50}<| {0:>3} %', end='')
+            transferred = 0
+            lastpercent = 0
             while True:
                 data = urlfile.read(32*1024) # large chunks plz
                 if not data:
-                    console(f"\rDownload {remote_url} finished.                 ")
+                    print(f'\r    |<{"="*50}| {percent:>3} %')
                     return local_file
                 output.write(data)
-                written += len(data)
-                progress = int((written*100)/total)
-                if (progress - prev_progress) >= 5: # report every 5%
-                    prev_progress = progress
-                    written_megas = int(written/(1024*1024))
-                    console(f"\rDownloading {remote_url} {written_megas}/{total_megas}MB ({progress}%)...")
+                transferred += len(data)
+                percent = int((transferred / size) * 100.0)
+                if abs(lastpercent - percent) >= 5: # report every 5%
+                    lastpercent = percent
+                    n = int(percent / 2)
+                    right = '=' * n
+                    left = ' ' * int(50 - n)
+                    print(f'\r    |{left}<{right}| {percent:>3} %', end='')
 
 
 def unzip(local_zip, extract_dir):
     with zipfile.ZipFile(local_zip, "r") as zip:
         zip.extractall(extract_dir)
-    console(f'Extracted {local_zip} to {extract_dir}')
 
 
-def download_and_unzip(remote_zip, extract_dir, unless_file_exists):
-    if unless_file_exists and os.path.exists(unless_file_exists):
-        console(f"Skipping {os.path.basename(remote_zip)} because {unless_file_exists} exists.")
-        return
-    local_file = download_file(remote_zip, extract_dir)
+def download_and_unzip(remote_file, extract_dir, local_file):
+    if local_file and os.path.exists(local_file):
+        console(f"Skipping {os.path.basename(remote_file)} because {local_file} exists.")
+        return extract_dir
+    local_file = download_file(remote_file, extract_dir)
+    if not local_file:
+        return None
     unzip(local_file, extract_dir)
+    console(f'Extracted {local_file} to {extract_dir}')
+    return extract_dir
 
 
 def _should_copy(src, dst):
@@ -231,13 +246,23 @@ def _should_copy(src, dst):
     return False
 
 
-def copy_file(src, dst):
-    if _should_copy(src, dst):
+def _passes_filter(src_file, filter):
+    if not filter: return True
+    for f in filter:
+        if src_file.endswith(f):
+            return True
+    return False
+
+
+def copy_file(src, dst, filter):
+    if _passes_filter(src, filter) and _should_copy(src, dst):
         #console(f'copy {src}\n --> {dst}')
         shutil.copy2(src, dst)
 
 
-def copy_dir(src_dir, out_dir):
+def copy_dir(src_dir, out_dir, filter=None):
+    if not os.path.exists(src_dir):
+        raise RuntimeError(f'copy_dir: {src_dir} does not exist!')
     root = os.path.dirname(src_dir)
     for fulldir, _, files in os.walk(src_dir):
         reldir = fulldir[len(root):].lstrip('\\/')
@@ -249,13 +274,13 @@ def copy_dir(src_dir, out_dir):
         for file in files:
             src_file = os.path.join(fulldir, file)
             dst_file = os.path.join(dst_folder, file)
-            copy_file(src_file, dst_file)
+            copy_file(src_file, dst_file, filter)
 
 
-def copy_if_needed(src, dst):
+def copy_if_needed(src, dst, filter=None):
     #console(f'COPY {src} --> {dst}')
     if os.path.isdir(src):
-        copy_dir(src, dst)
+        copy_dir(src, dst, filter)
     else:
-        copy_file(src, dst)
+        copy_file(src, dst, filter)
 
